""" contains the classes BasisElement, 
    Element, and ArcAlgebra. they should 
    all work..
"""


class BasisElement:
    """ basis element of Hn/ArcAlgebra """

    def __init__(self, cm1, cm2, loops, k=0, inlabels=None):
        """ inputs: two crossingless matchings, an accounting of 
                their circles, and (an integer whose base 2 
                representation corresponds to the labelling of 
                the circles -or- a list of labels).
            checks inputs to make sure they make sense
        """
        if len(cm1) != len(cm2):
            raise Exception("matchings not the same length")
        if not isxless(cm1):
            raise Exception("first matching not crossingless")
        if not isxless(cm2):
            raise Exception("second matching not crossingless")
        
        self.cm1 = tuple(cm1)
        self.cm2 = tuple(cm2)
        self.loops = tuple(loops)

        if inlabels == None:
            if k > 2**len(loops):
                raise Exception("invalid labelling")
            labels = [0]*len(loops) 
            cc = len(loops) #cc for circle count
            counter = k
            for i in range(cc-1, -1, -1):
                if counter >= 2**i:
                    counter -= 2**i
                    labels[cc-1-i] = 1 
            self.labels = tuple(labels)
        else:
            if len(inlabels) != len(loops):
                raise Exception("given labellings of the wrong length")
            self.labels = tuple(inlabels) 

    def __repr__(self):
        s = str(self.cm1)
        s += str(self.cm2)
        s += str(self.loops)
        s += str(self.labels)
        return s
    
    def __eq__(self, other):
        if isinstance(other, BasisElement):
            return (self.cm1 == other.cm1) and (self.cm2 == other.cm2) \
                and (self.labels == other.labels)
        if isinstance(other, Element):
            if self in other.dict and other.dict[self] == 1:
                for key in other.dict:
                    if key != self and other.dict[key] != 0:
                        return False
                return True
        return False
    
    def __hash__(self):
        # cm1 and cm2 determine loops, so we only need these 3.
        bigtuple = self.cm1 + self.cm2 + self.labels
        return hash(bigtuple)

    def mult(self, other, returnElt=True):
        """ multiplies two BasisElements together, self*other.
            returns an Element unless returnElt is False, 
            when it returns the corresponding 
            dictionary instead. (for Element's mult.)
        """
        #RL: should check here if idempotents agree and immediately return 0 if not.
        scomps = self.get_comps()
        ocomps = other.get_comps()
        merged_dict = dict()
        for comp in scomps.keys():
            newc = tuple( ((a,b,'1'+c,d) for (a,b,c,d) in comp))
            merged_dict[newc] = scomps[comp]
        for comp in ocomps.keys():
            newc = tuple( ((a,b,'2'+c,d) for (a,b,c,d) in comp) )
            merged_dict[newc] = ocomps[comp]
        merged_comps = self._nm_recurse(merged_dict)
        #drop the 1s and 2s
        merged_elts = list()
        for elt in merged_comps:
            fixed_elt = dict()
            for comp in elt.keys():
                newc = tuple( ((a,b,c[1],d) for (a,b,c,d) in comp))
                fixed_elt[newc] = elt[comp]
            merged_elts.append(fixed_elt)
        answer = Element([_comps_to_elt(x) for x in merged_elts], len(merged_elts)*[1,])
        if returnElt:
            return answer
        else:
            return answer.dict
        
    def _nm_recurse(self,merged_dict):
        "Helper function for recursion in new_mult."
        #Find an unmerged strand. This is arguably inefficient.
        found = False
        for c in merged_dict.keys():
            for s in c:
                if s[2]=='1r':
                    found = True
                    break
            if found:
                break
        if not found:
            return [merged_dict,]        #If there wasn't one, return a list with merged_dict as the only element
        (a,b,blah,d) = s
        if ( (a,b,'2l',0) in c) or ( (a,b,'2l',1) in c):         #Is the unmerged strand a split, i.e., are there two copies on the same component?
            if (a,b,'2l',0) in c:
                t = (a,b,'2l',0)
            else:
                t = (a,b,'2l',1)
            #Find result of splitting this component
            i = c.index(s)
            j = c.index(t)
            m = min(i,j)
            n = max(i,j)
            c1 = c[m+1:n]
            c2 = c[n+1:]+c[:m]
            #Copy over the rest of the components
            new_dict = merged_dict.copy()
            new_dict.pop(c)
            if merged_dict[c] == 1: #The circle is labeled X
                new_dict[c1] = 1
                new_dict[c2] = 1
                return self._nm_recurse(new_dict)        #One  term in split, recurse on it, return result.
            #The circle is labeled 1
            nd2 = new_dict.copy()
            new_dict[c1] = 0
            new_dict[c2] = 1
            nd2[c1] = 1
            nd2[c2] = 0
            return self._nm_recurse(new_dict)+self._nm_recurse(nd2)        #Two terms in split, recurce on both, return sum of two lists.

        #Not a split, so it's a merge. Find the other copy.
        for d in merged_dict.keys():
            if ( (a,b,'2l',0) in d ) or ( (a,b,'2l',1) in d ):
                break
        if merged_dict[c] == 1 and merged_dict[d] == 1:         #If the product is zero, return empty list.
            return list()
        #Copy over the rest of the components
        new_dict = merged_dict.copy()
        new_dict.pop(c)
        new_dict.pop(d)
        #Build the new component
        i = c.index(s)
        try:
            j = d.index((a,b,'2l',0))
            up = 0
        except:
            j = d.index((a,b,'2l',1))
            up = 1
        if s[3] == up: #Both circles oriented the same way; need to reverse the latter.
            newd = list()
            for k in range(len(d)):
                (alph,bet,gam,delt) = d[len(d)-k-1]
                newd.append((alph,bet,gam,1-delt))
            newd = tuple(newd)
            j = len(newd) - j-1
            assert newd[j][0] == a and newd[j][1] == b
        else:
            newd = d
        #Now we can merge
        merged_circ = c[:i]+newd[j+1:]+newd[:j]+c[i+1:]
        new_dict[merged_circ] = merged_dict[c]+merged_dict[d]
        return self._nm_recurse(new_dict)#Recurse on the product and return the result.
        

    def get_comps(self):
        """Returns a dictionary with keys connected components and values 1 or X.
        Connected components are stored as follows. For the crossingless matching (3,2,1,0) 
        on the left and (1,0,3,2) on the right, the connected component is stored as
        ((0,3,'l',0),(2,3,'r',1),(1,2,'l',1),(0,1,'r',1))
        The first two entries are the points connected by the segment. The 'l' or 'r' 
        indicates if the segment is on the left or the right. The last entry indicates if the 
        segment is going from the first entry to the second (0) or the second entry to the first (1).

        The result is not canonical: there is an action of a dihedral
        group on each component by cyclic permutation of the intervals
        and reversing the orientation of the component.
        """
        lnextdict = dict()
        rnextdict = dict()
        complist = list()
        n = len(self.cm1)
        touse = list(range(n))
        while touse:
            thiscomp = list()
            compstart = touse[0]
            touse.remove(compstart)
            here = self.cm1[compstart]
            touse.remove(here)
            if here > compstart:
                thiscomp.append((compstart,here,'l',0))
            else:
                thiscomp.append((here,compstart,'l',1))
            there = self.cm2[here]
            if there > here:
                thiscomp.append((here,there,'r',0))
            else:
                thiscomp.append((there,here,'r',1))
            while there != compstart:
                touse.remove(there)
                here = there
                there = self.cm1[here]
                if there > here:
                    thiscomp.append((here,there,'l',0))
                else:
                    thiscomp.append((there,here,'l',1))
                touse.remove(there)
                here = there
                there = self.cm2[here]
                if there > here:
                    thiscomp.append((here,there,'r',0))
                else:
                    thiscomp.append((there,here,'r',1))
            complist.append(thiscomp)
        answer = dict()
        for comp in complist:
            p = comp[0][0]
            for (i,l) in enumerate(self.loops):
                if p in l:
                    answer[tuple(comp)] = self.labels[i]
                    break
        return answer
    

    def old_mult(self, other, returnElt=True):
        """ multiplies two BasisElements together, self*other.
            returns an Element unless returnElt is False, 
            when it returns the corresponding 
            dictionary instead. (for Element's mult.)
        """
        if self.cm2 != other.cm1:
            if returnElt:
                return Element(self, coeff = 0)
            else:
                return {self: 0}

        twoN = len(self.cm2)
        productLoops = loops(self.cm1, other.cm2)
        ms = matchStarts(self.cm2)
        loopsInProgress = [[list(loop) for loop in self.loops] + \
            [[i+twoN for i in loop] for loop in other.loops], \
            [[val] for val in self.labels] + [[val] for val in other.labels] ]

        for i in ms: # one merge/split per matched pair
            loop1, label1, loopsInProgress = \
                removeLoopContaining(i,loopsInProgress)
            if (self.cm2[i] + twoN) in loop1:
                # split case:
                cycleInd = loop1.index(self.cm2[i])
                loop1 = loop1[cycleInd:] + loop1[:cycleInd]
                splitInd = loop1.index(self.cm2[i]+twoN)
                newLoop1 = loop1[:splitInd+1]
                newLoop2 = loop1[splitInd+1:]
                newLabel1, newLabel2 = split(label1)
                loopsInProgress[0] += [newLoop1]
                loopsInProgress[1] += [newLabel1] 
                loopsInProgress[0] += [newLoop2]
                loopsInProgress[1] += [newLabel2]
            else: 
                # merge case!
                loop2, label2, loopsInProgress = \
                    removeLoopContaining(self.cm2[i]+twoN,loopsInProgress)
                newLoop = loop1 + loop2
                loopsInProgress[0] += [newLoop]
                loopsInProgress[1] += [merge(label1, label2)]
        if returnElt:
            return Element(dictionary=makeDict(self.cm1,other.cm2, 
                productLoops,loopsInProgress))
        else:
            return makeDict(self.cm1, other.cm2, productLoops, 
                loopsInProgress)
        return None
    
    def getK(self):
        """ gets the value k back from the 
            BasisElement; just to help w 
            the capping off case in chaincx
        """
        k = 0
        for i in range(1,len(self.labels)+1):
            k += self.labels[-i]*2**(i-1)
        return k

def _comps_to_elt(comps):
    "Take a dictionary of components, like in get_comps, and return a BasisElement."
    l_matching_dict = dict()
    r_matching_dict = dict()
    for comp in comps.keys():
        for (a,b,c,d) in comp:
            if c == 'l':
                l_matching_dict[a] = b
                l_matching_dict[b] = a
            else:
                r_matching_dict[a] = b
                r_matching_dict[b] = a
    n = len(l_matching_dict.keys())
    assert len(l_matching_dict.keys()) == len(r_matching_dict.keys())
    l_matching = tuple((l_matching_dict[i] for i in range(n)))
    r_matching = tuple((r_matching_dict[i] for i in range(n)))
    circs = loops(l_matching,r_matching)
    labels = list()
    for i in range(len(circs)):
        (alph,bet) = (circs[i][0],circs[i][1])
        a = min(alph,bet)
        b = max(alph,bet)
        for comp in comps.keys():
            if ( (a,b,'l',0) in comp ) or ( (a,b,'l',1) in comp ):
                labels.append(comps[comp])
                break
    return BasisElement(l_matching,r_matching,circs,inlabels=labels)
    
class Element:
    """ arbitrary element of Hn/ArcAlgebra """

    def __init__(self, BasisElts=None, coeff = 1, dictionary = None):
        """ initializer. can take in:
            - a single basis element and a single coefficient 
            - a list of basis elements and corresponding list of coefficients
            - a dictionary of basis elements and coefficients.
            to get a zero just use the first option with any basis 
            element of your choice and zero as the coeffient.
            the empty dictionary might also be zero..
        """
        if BasisElts != None:
            self.dict = {}
            if isinstance(BasisElts, list):
                if len(BasisElts) != len(coeff):
                    raise Exception("mismatched inputs")
                for i in range(len(BasisElts)):
                    if not BasisElts[i] in self.dict.keys():
                        self.dict[BasisElts[i]] = coeff[i]
                    else:
                        self.dict[BasisElts[i]] += coeff[i]
            else:
                self.dict = { BasisElts : coeff}
        elif dictionary != None:
            self.dict = dictionary
        else:
            # if you just call Element() you get zero!!
            # good work figuring out what you want it to be :)
            self.dict = {}

    def __repr__(self):
        s = str(self.dict)
        return s
    
    def __eq__(self, other):
        """ standard equality operator, does what one might expect.
            will say that two elements are equal if both are zero 
            but in different representations. we can also compare 
            Elements and BasisElements.
        """
        if isinstance(other, BasisElement):
            if other in self.dict and self.dict[other] == 1:
                for key in self.dict:
                    if key != other and self.dict[key] != 0:
                        return False
                return True
            else:
                return False
        
        if self.dict == other.dict:
            return True
        else:
            for key in self.dict:
                if self.dict[key] != 0:
                    if key not in other.dict or \
                        other.dict[key] != self.dict[key]:
                        return False
            for key in other.dict:
                if other.dict[key] != 0:
                    if key not in self.dict:
                        return False
        return True

    def __add__(self, other):
        if isinstance(other, Element):
            return self.add(other)
        if isinstance(other, BasisElement):
            return self+Element(other)
        raise(NotImplementedError)

    def __mul__(self, other):
        if isinstance(other, int):
            newdict = {}
            for x in self.dict:
                newdict[x] = other*self.dict[x]
            return Element(dictionary=newdict)
        if isinstance(other, Element):
            return self.mult(other)
        if other.__class__ == dict:
            return self.mult(Element(dictionary=other))            
        return other.__rmul__(self)
    
    
    def mult(self, other):
        #RL: This is self*other
        newdict = {}
        for elt, coeff in self.dict.items():
            for elt2, coeff2 in other.dict.items():
                if elt.cm2 != elt2.cm1:
                    # if they are not Compatible we simply move on.
                    continue
                # multiply elt and elt2 in BasisElement to get a dictionary.
                result = elt.mult(elt2, False)
                # add to our running result
                for eltR, coeffR in result.items():
                    if eltR in newdict:
                        newdict[eltR] += coeff*coeff2*coeffR
                    else:
                        newdict[eltR] = coeff*coeff2*coeffR
        return Element(dictionary=newdict)

    def addInPlace(self, other):
        """ adding in place """
        """ not recommended for use because it sometimes does weird things. 
            dunno if that's gonna be a Problem..
            don't use. if you can help it..
        """
        for key in self.dict:
            if key in other.dict:
                self.dict[key] += other.dict[key]
        for key in other.dict:
            if key not in self.dict:
                self.dict[key] = other.dict[key]
                
    def add(self,other):
        """ adding not in place (i.e. makes new Element)
        """
        newdict = {}
        for key in self.dict:
            if key in other.dict:
                c = self.dict[key] + other.dict[key]
                if c != 0: #Don't include terms with coefficient 0.
                    newdict[key] = self.dict[key] + other.dict[key]
            else:
                newdict[key] = self.dict[key]
        for key in other.dict:
            if key not in self.dict:
                newdict[key] = other.dict[key]
        return Element(dictionary=newdict)
    
    def scalarMult(self,scalar):
        """ multiplies each coefficient 
            by the given scalar.
        """
        for key in self.dict:
            self.dict[key] *= scalar

class ArcAlgebra:
    """ big ring class """

    def __init__(self, n):
        self.n = n 
        self.basis = []
        # create list of xless matchings
        lxm = xless(n)
        # double loop through them
        for i in lxm:
            for j in lxm:
                circles = loops(i,j)
                for k in range(2**len(circles)):
                    # all possible labelings.
                    self.basis += [BasisElement(i, j, circles, k)]

    def getBasis(self):
        return self.basis

def idempotent_basis_elt(matching):
    "Returns the BasisElement corresponding to (matching,matching) with all circles labeled 1 (an indecomposable idempotent in the algebra)"
    the_loops = loops(matching,matching)
    return BasisElement(matching, matching, the_loops)

def idempotent(matching):
    "Returns the Element corresponding to (matching,matching) with all circles labeled 1 (an indecomposable idempotent in the algebra)"
    ibe = idempotent_basis_elt(matching)
    return Element(BasisElts=ibe,coeff=1)


# Helper functions!!
def xless(n):
    """ takes in an integer n and returns 
        all crossingless matchings of 2n points
        in the form of a list of lists. 
    """
    if n == 1:
        return [[1,0]]
    oldlist = xless(n-1)
    newlist = []
    for matching in oldlist:
        a = matching + [2*n - 1, 2*n - 2]
        if a not in newlist:
            newlist += [a]
        b = [1,0] + [x+2 for x in matching]
        if b not in newlist:
            newlist += [b]
        newlist += [[2*n - 1] + [x+1 for x in matching] + [0]]
    return newlist
    
def loops(cm1, cm2):
    """ returns tuple of tuples of indices 
        in each loop of the pairing cm1(mirrored) cm2.
    """
    lol = []
    for i in range(len(cm1)):
        if (i > 0) and \
            (any(i in loop for loop in lol)):
            continue
        newloop = []
        current = i
        while True:
            newloop += [current]
            current = cm1[current]
            newloop += [current]
            current = cm2[current]
            if current == i:
                break
        lol += [tuple(newloop)]
    return tuple(lol)

def isxless(matching):
    """ checks if a matching (in the form of 
        a tuple/list) is crossingless
    """ 
    for i in range(len(matching)):
        if (i - matching[i]) % 2 == 0:
            # a faster test that doesn't catch everything.
            return False
        else: # the more precise but n^2 test.
            for j in range(i):
                if not (matching[j] < i) and \
                    not (matching[j] > matching[i]):
                    return False
    return True

def merge(a,b):
    """ the merge map; 
        0 means 1, 1 means x;
        (1,0) and (0,1) go to 1 (x)
        (0,0) goes to 0 (1),
        (1,1) goes to -1 (0).
        (-1,anything) goes to -1
    """
    merged = [0]*len(a)
    for i in range(len(a)):
        if a[i] == b[i]:
            if a[i] == 1:
                merged[i] = -1
            elif a[i] == 0:
                merged[i] = 0
            elif a[i] == -1:
                merged[i] = -1
        elif a[i] == -1 or b[i] == -1:
            merged[i] = -1
        else:
            merged[i] = 1
    return merged

def split(a):
    """ the split map;
        takes 1 to (1,1)
        and takes 0 to (1,0) + (0,1)
    """
    split1 = [0]*len(a)
    split2 = [0]*len(a)
    for i in range(len(a)):
        if a[i] == 1:
            split1[i] = 1
            split2[i] = 1
        elif a[i] == 0:
            split1[i] = [0,1]
            split2[i] = [1,0]
        elif a[i] == -1:
            split1[i] = -1
            split2[i] = -1
        elif isinstance(a[i], list):
            subsplit1, subsplit2 = split(a[i])
            split1[i] = subsplit1
            split2[i] = subsplit2       
    return split1, split2

def matchStarts(cm):
    """ given a crossingless matching, returns 
        a list of the indices at which 
        the matchings start.
    """
    starts = []
    for i in range(len(cm)):
        if cm[i] > i:
            starts += [i]
    return starts

def removeLoopContaining(elt, inlist):
    """ given a list of lists (inlist), 
        removes the sublist containing elt, 
        and cycles the desired sublist to 
        begin with elt
    """
    outlist = []
    wanted = []
    for sublist in inlist[0]:
        if elt in sublist:
            i = sublist.index(elt)
            wanted = sublist[i:] + sublist[:i]
            j = inlist[0].index(sublist)
            label = inlist[1][j]
            newlabels = inlist[1][:j] + inlist[1][j+1:]
        else:
            outlist += [sublist]
    return wanted, label, [outlist, newlabels]

def makeDict(cm1, cm2, loops, LoL):
    """ makes a nice dictionary of labelled
        BasisElements w coefficients
    """ 
    dictionary = {}
    LoL = reorder(loops, LoL)
    listOfLabels = flatten(LoL[1])
    for elt in listOfLabels:
        basisElt = BasisElement(cm1,cm2,loops,inlabels=elt)
        if basisElt in dictionary:
            dictionary[basisElt] += 1
        else:
            dictionary[basisElt] = 1
    return dictionary

def flatten(LoL):
    """ given labelling data sorted by 
        loops, converts it to a 
        collection of labellings of a 
        diagram
    """
    output = []
    for i in range(len(LoL[0])):
        templist = [0]*len(LoL)
        for j in range(len(LoL)):
            if LoL[j][i] == -1:
                # since that Means 0.
                break
            templist[j] = LoL[j][i]
            if j == len(LoL) - 1:
                # so that we only add when we don't break out.
                output += splitlist(templist)
    return output # a list of labelllings.

def reorder(order, LoL):
    """ reorders loopsInProgress so that 
        the labels are in the correct order
        (because flatten assumes this)
    """
    output = [[],[]]
    for loop in order:
        for index in range(len(LoL[0])):
            if loop[0] in LoL[0][index]:
                output[0] += [LoL[0][index]]
                output[1] += [LoL[1][index]]
                break
    return output

def splitlist(liist):
    """ takes in a list of possibly lists,
        splits so that each nonlist element 
        is the same in the outputs, but list 
        elements are either first or second
        (helper function for flatten)
    """
    list1 = [0]*len(liist)
    list2 = [0]*len(liist)
    for i in range(len(liist)):
        if isinstance(liist[i],list):
            list1[i] = liist[i][0]
            list2[i] = liist[i][1]
        else:
            list1[i] = liist[i]
            list2[i] = liist[i]
    if list1 == list2:
        return [list1]
    return splitlist(list1) + splitlist(list2)

# # some small examples!
# basis3 = ArcAlgebra(3).basis
# x = basis3[8]
# y = basis3[22]
# z = basis3[46]
# w = basis3[80]
# print(x.mult(y))
# print(z.mult(w))
# a = Element(x)
# zero = Element(dictionary={})
# b = Element([x,y],[3,4])
# print(b.mult(a))


#The element zero
E0 = Element(dictionary=dict())
