import unittest
from arcalg import idempotent, E0, Element, BasisElement

class TestIdempotents(unittest.TestCase):
    def test_0(self):
        self.assertEqual(E0, E0+E0)
        self.assertEqual(E0, E0*E0)
    def test_idems(self):
        i = idempotent((3,2,1,0))
        j = idempotent((1,0,3,2))
        self.assertEqual(i, i*i)
        self.assertEqual(j, j*j)
        self.assertEqual(i*j, E0)
        self.assertEqual(j*i, E0)
        self.assertEqual((i+j)*(i+j),i+j)
        self.assertFalse(i==E0)
        self.assertFalse(i==j)

class TestAlgMul(unittest.TestCase):
        def test_elt_mull(self):
            i = idempotent((3,2,1,0))
            j = idempotent((1,0,3,2))
            a = BasisElement((1,0,3,2),(1,0,3,2),[(0,1),(2,3)],inlabels=(1,0))
            b = BasisElement((1,0,3,2),(1,0,3,2),[(0,1),(2,3)],inlabels=(0,1))
            c = BasisElement((1,0,3,2),(1,0,3,2),[(0,1),(2,3)],inlabels=(1,1))
            d = BasisElement((3,2,1,0),(1,0,3,2),[(0,1,2,3)],inlabels=(0,))
            e = BasisElement((1,0,3,2),(3,2,1,0),[(0,1,2,3)],inlabels=(0,))
            A = Element(a)
            B = Element(b)
            C = Element(c)
            D = Element(d)
            E = Element(e)
            self.assertEqual(j*A,A)
            self.assertEqual(i*A,E0)
            self.assertEqual(i*D*j,D)
            self.assertEqual(A*B,C)
            self.assertFalse(D*E==E*D)
            self.assertEqual(E*D,A+B)

        def more_basis_mul_tests(self):
            a = BasisElement((5,4,3,2,1,0),(1,0,5,4,3,2),loops((5,4,3,2,1,0),(1,0,5,4,3,2)),inlabels=(0,))
            b = BasisElement((1,0,5,4,3,2),(3,2,1,0,5,4),loops((1,0,5,4,3,2),(3,2,1,0,5,4)),inlabels=(0,))
            x = a.mult(b)
            self.assertEqual(list(x.dict.values()),[2,])

        def test_dict_mul(self):
            a = BasisElement((1,0,3,2),(1,0,3,2),[(0,1),(2,3)],inlabels=(1,0))
            b = BasisElement((1,0,3,2),(1,0,3,2),[(0,1),(2,3)],inlabels=(0,1))
            c = BasisElement((1,0,3,2),(1,0,3,2),[(0,1),(2,3)],inlabels=(1,1))
            A = Element(a)
            B = {b:1}
            C = Element(c)
            self.assertEqual(A*B,C)
        
if __name__ == '__main__':
    unittest.main()
