import unittest
from arcalg import BasisElement, Element, idempotent
from khcx import ModuleElement, ChainCx,  closed_braid

class TestModuleElement(unittest.TestCase):
    def test_add(self):
        d = Element(BasisElement((3,2,1,0),(1,0,3,2),[(0,1,2,3)],inlabels=(0,)))
        c = Element(BasisElement((1,0,3,2),(1,0,3,2),[(0,1),(2,3)],inlabels=(1,1)))
        x =  ModuleElement({'0':c})
        y =  ModuleElement({'0':d})
        z = ModuleElement({'0':(c+d)})
        self.assertEqual(x+y,z)
        u =  ModuleElement({'0':c})
        v =  ModuleElement({'1':d})
        w =  ModuleElement({'0':c,'1':d})
        self.assertEqual(u+v,w)

    def test_mul(self):
        d = Element(BasisElement((3,2,1,0),(1,0,3,2),[(0,1,2,3)],inlabels=(0,)))
        c = Element(BasisElement((1,0,3,2),(1,0,3,2),[(0,1),(2,3)],inlabels=(1,1)))
        i = idempotent((3,2,1,0))
        j = idempotent((1,0,3,2))
        x =  ModuleElement({'0':j,'1':i})
        y = ModuleElement({'0':d})
        self.assertEqual(d*x,y)
        self.assertFalse(x==y)

class TestDSquared(unittest.TestCase):
    def testDSquareD(self):
        tangle1 = ChainCx([('d',0,4),('u',1,4),(1,0,3,2)], simplify=False)
        self.assertTrue(tangle1.test_d_squared())
        tref1 = ChainCx([(1,0,3,2),('d',1,4),('u',2,4),('u',0,4),(3,2,1,0)])
        self.assertTrue(tref1.test_d_squared())
        
class TestRationalHomology(unittest.TestCase):
    def testTrefoil(self):
        tref1 = ChainCx([(1,0,3,2),('d',1,4),('u',2,4),('u',0,4),(3,2,1,0)])
        self.assertEqual(tref1.rational_homology(), 4)

    def testTrefBraids(self):
        K3_1 = closed_braid([1, 1, 1]) #RHT
        mK3_1 = closed_braid([-1, -1, -1]) #LHT
        self.assertEqual(K3_1.rational_homology(), 4)
        self.assertEqual(mK3_1.rational_homology(), 4)

    def testFig8(self):
        K4_1 = closed_braid([1, -2, 1, -2])
        self.assertTrue(K4_1.test_d_squared())
        self.assertEqual(K4_1.rational_homology(), 6) #From knot atlas

    def testMoreBraids(self):
        K5_2 = closed_braid([-1, 2, -3, 2, 2, 1, 2, 3, 2], simplify=False)
        K5_2simp = closed_braid([-1, 2, -3, 2, 2, 1, 2, 3, 2])
        K8_19 = closed_braid([1,2,1,2,1,2,1,2]) #T(3,4)
        K9_42 = closed_braid([-1,-1,-1,2,1,1,-3,2,-3])
#        K9_42 = closed_braid([-1, 2, 2, 2, -3, -2, -2, 1, 2, -3, 2])
        #K9_46 = closed_braid([-1, 2, -3, -4, -5, -4, 6, 3, 7, -2, 1, 3, 2, 4, 3, -4, -5, -4, -6, -3, 5, -2, 4, -3, 4, -5, -4, 6, 5, -7, -6, 5, -4])
        self.assertTrue(K5_2.test_d_squared())
        self.assertTrue(K5_2simp.test_d_squared())
        self.assertEqual(K5_2simp.rational_homology(), 8) #From knot atlas
        self.assertEqual(K8_19.rational_homology(), 8) #From knot atlas 
        self.assertEqual(K9_42.rational_homology(), 10) #From knotinfo
#        self.assertEqual(K9_46.rational_homology(), 10) #From knot atlas

class TestGradings(unittest.TestCase):
    def testBraidGradingsNonClosed(self):
        K3_1op = closed_braid([1, 1, 1], simplify=False, closed=False) #RHT
        mK3_1op = closed_braid([-1, -1, -1], simplify=False, closed=False) #LHT
        K4_1op = closed_braid([1, -2, 1, -2], simplify=False, closed=False)
        self.assertTrue(K3_1op.test_diff_gr())
        self.assertTrue(mK3_1op.test_diff_gr())
        self.assertTrue(K4_1op.test_diff_gr())
        
    def testBraidGradingsClosed(self):
        K3_1 = closed_braid([1, 1, 1], simplify=False) #RHT
        K5_2 = closed_braid([-1, 2, -3, 2, 2, 1, 2, 3, 2],simplify=False)
        K8_19 = closed_braid([1,2,1,2,1,2,1,2]) #T(3,4)
        self.assertTrue(K3_1.test_diff_gr())
        self.assertTrue(K5_2.test_diff_gr())
        self.assertTrue(K8_19.test_diff_gr())
        
if __name__ == '__main__':
    unittest.main()
